/* yalias: a GTK command launcher.                    vim:et:ts=2
   Copyright 2003 Adam Sampson <azz@us-lot.org>

   yalias is free software; you can redistribute and/or modify it
   under the terms of that license as published by the Free Software
   Foundation; either version 2 of the License, or (at your option)
   any later version.
   
   yalias is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with yalias; see the file COPYING. If not, write to the
   Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307 USA, or see http://www.gnu.org/.
 */

#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <gdk/gdkkeysyms.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <sys/types.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <regex.h>
#include <unistd.h>

GtkWidget *window, *hbox, *entry, *clear, *clear_image, *execute,
  *execute_image;
guchar *last = NULL;

typedef struct {
  regex_t re;
  char *command;
} def;

#define MAXDEFS 100
def defs[MAXDEFS];
int num_defs = 0;

char *get_match(const char *buf, regmatch_t *match) {
  int l = match->rm_eo - match->rm_so;
  char *ret = malloc(l + 1);

  strncpy(ret, buf + match->rm_so, l);
  ret[l] = '\0';

  return ret;
}

void read_config(char *name) {
  FILE *f = fopen(name, "r");
  char buf[1024];
  regex_t re;
  regmatch_t match[3];

  if (!f) {
    fprintf(stderr, "Unable to read config file %s\n", name);
    exit(20);
  }

  regcomp(&re, "^([^=]*)=(.*)$", REG_EXTENDED);

  while (fgets(buf, sizeof buf, f) != NULL && num_defs < MAXDEFS) {
    int rc;

    buf[sizeof buf - 1] = '\0';
    buf[strlen(buf) - 1] = '\0';
    rc = regexec(&re, buf, 3, match, 0);
    if (rc == REG_NOMATCH || match[1].rm_so < 0 || match[2].rm_so < 0) {
      fprintf(stderr, "Unrecognised line '%s'\n", buf);
      exit(20);
    } else {
      char *ms = get_match(buf, &match[1]);

      if (regcomp(&defs[num_defs].re, ms, REG_EXTENDED) != 0) {
        fprintf(stderr, "Bad regexp '%s'\n", ms);
        exit(20);
      }
      free(ms);
      defs[num_defs].command = get_match(buf, &match[2]);

      ++num_defs;
    }
  }

  fclose(f);
}

gboolean handle_delete(GtkWidget *widget, GdkEvent *event, gpointer data) {
  return FALSE;
}

void handle_destroy(GtkWidget *widget, gpointer data) {
  gtk_main_quit();
}

void handle_clear(GtkWidget *widget, gpointer data) {
  gtk_entry_set_text(GTK_ENTRY(entry), "");
}

void reset_window() {
  gtk_entry_set_text(GTK_ENTRY(entry), "");
  gtk_widget_grab_focus(entry);
}

void run_command(char *cmd) {
  GError *error = NULL;
  gchar *argv[4] = { "/bin/sh", "-c", cmd, NULL };

  g_spawn_async(NULL, argv, NULL, 0, NULL, NULL, NULL, &error);
}

void handle_execute(GtkWidget *widget, gpointer data) {
  int i;
  const char *text = gtk_entry_get_text(GTK_ENTRY(entry));
  char *p, *q;
  char buf[1024];
#define MAXMATCHES 10
  regmatch_t match[MAXMATCHES];

  for (i = 0; i < num_defs; i++) {
    if (regexec(&defs[i].re, text, MAXMATCHES, match, 0) != REG_NOMATCH)
      goto found;
  }
  return;

found:
  p = defs[i].command;
  q = buf;

  /* Eurgh. TODO: port this to use iolib. */
  while (q < (buf + sizeof buf - 1)) {
    char c = *p++;
    if (c == '\0') {
      break;
    } else if (c == '$') {
      int i, n;
      char *sub;

      c = *p++;
      if (c == '\0') break;

      i = c - '0';
      if (i < 0 || i > 9 || match[i].rm_so < 0) {
        fprintf(stderr, "Bad $n in command '%s'\n", defs[i].command);
        return;
      }

      sub = get_match(text, &match[i]);
      n = strlen(sub);
      if (q + n > (buf + sizeof buf - 1)) break;
      strcpy(q, sub);
      q += n;
    } else {
      *q++ = c;
    }
  }
  *q = '\0';

  run_command(buf);
  reset_window();
}

int main(int argc, char **argv) {
  GtkAccelGroup *accel_group;
  const char *home = getenv("HOME");
  char buf[1024];
  const char *geom = NULL;
  int buttons = 1;
  int hint = 0;
  int stick = 0;

  gtk_init(&argc, &argv);
  snprintf(buf, sizeof buf, "%s/.yaliasrc", home ? home : ".");

  while (1) {
    int c = getopt(argc, argv, "c:hbsg:");
    if (c == -1) break;

    switch (c) {
    case 'c':
      snprintf(buf, sizeof buf, "%s", optarg);
      break;
    case 'b':
      buttons = 0;
      break;
    case 'h':
      hint = 1;
      break;
    case 's':
      stick = 1;
      break;
    case 'g':
      geom = strdup(optarg);
      break;
    default:
      fprintf(stderr, "Usage: yalias [-h] [-b] [-s] "
                      "[-g geometry] [-c config-file]\n");
      return 20;
    }
  }

  read_config(buf);

  entry = gtk_entry_new();

  accel_group = gtk_accel_group_new();

  clear_image = gtk_image_new_from_stock(GTK_STOCK_CLEAR,
                                         GTK_ICON_SIZE_BUTTON);

  clear = gtk_button_new();
  g_signal_connect(G_OBJECT(clear), "clicked",
                   G_CALLBACK(handle_clear), NULL);
  gtk_widget_add_accelerator(clear, "clicked", accel_group,
                             GDK_Escape, 0, 0);
  gtk_container_add(GTK_CONTAINER(clear), clear_image);

  execute_image = gtk_image_new_from_stock(GTK_STOCK_EXECUTE,
                                           GTK_ICON_SIZE_BUTTON);

  execute = gtk_button_new();
  g_signal_connect(G_OBJECT(execute), "clicked",
                   G_CALLBACK(handle_execute), NULL);
  gtk_widget_add_accelerator(execute, "clicked", accel_group,
                             GDK_Return, 0, 0);
  gtk_container_add(GTK_CONTAINER(execute), execute_image);

  hbox = gtk_hbox_new(FALSE, 0);
  if (buttons) gtk_box_pack_start(GTK_BOX(hbox), clear, TRUE, TRUE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
  if (buttons) gtk_box_pack_end(GTK_BOX(hbox), execute, TRUE, TRUE, 0);
  gtk_widget_show_all(hbox);

  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(window), "yalias");
  gtk_window_set_resizable(GTK_WINDOW(window), FALSE);
  g_signal_connect(G_OBJECT(window), "delete_event",
                   G_CALLBACK(handle_delete), NULL);
  g_signal_connect(G_OBJECT(window), "destroy",
                   G_CALLBACK(handle_destroy), NULL);
  gtk_window_add_accel_group(GTK_WINDOW(window), accel_group);
  gtk_container_add(GTK_CONTAINER(window), hbox);
  if (geom != NULL) {
    gtk_window_parse_geometry(GTK_WINDOW(window), geom);
  }
  gtk_widget_realize(window);
  if (hint) {
    gtk_window_set_type_hint(GTK_WINDOW(window), GDK_WINDOW_TYPE_HINT_DOCK);
  }
  if (stick) gtk_window_stick(GTK_WINDOW(window));
  gtk_widget_show_all(window);

  reset_window();

  gtk_main();

  exit(0);
}

